/*
    UHU-menu
    Copyright (C) 2003  Koblinger Egmont <egmont@uhulinux.hu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#define _GNU_SOURCE
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <errno.h>
#include <stdarg.h>
#include "uhu-menu.h"

struct language *first_language = NULL;

static const char *try_locales[] = {
#include "locales.h"
	NULL
};

/*
 * newlocale2:
 * if name=it, then try it, it.UTF-8, it_IT, it_IT.UTF-8, it_CH, it_CH.UTF-8 in this order
 * etc...
 */

locale_t newlocale2 (const char *name)
{
	int i, l;
	locale_t new;
	char lang[32];
	char buf[64];
	char *p;

	p = strchr(name, '.');
	if (p != NULL) {
		strncpy(lang, name, p - name);
		lang[p - name] = '\0';
	} else {
		strcpy(lang, name);
	}

	sprintf(buf, "%s", lang);
	new = newlocale(LC_ALL_MASK, buf, NULL);
	if (new != NULL) {
		if (!strcmp(nl_langinfo_l(CODESET, new), "UTF-8")) return new;
		freelocale(new);
	}
	sprintf(buf, "%s.UTF-8", lang);
	new = newlocale(LC_ALL_MASK, buf, NULL);
	if (new != NULL) {
		if (!strcmp(nl_langinfo_l(CODESET, new), "UTF-8")) return new;
		freelocale(new);
	}

	if (strchr(name, '_') != NULL) return NULL;
	l = strlen(lang);
	i = 0;
	while (try_locales[i] != NULL) {
		if (!strncmp(try_locales[i], lang, l) && try_locales[i][l] == '_') {
			sprintf(buf, "%s", try_locales[i]);
			new = newlocale(LC_ALL_MASK, buf, NULL);
			if (new != NULL) {
				if (!strcmp(nl_langinfo_l(CODESET, new), "UTF-8")) return new;
				freelocale(new);
			}
			sprintf(buf, "%s.UTF-8", try_locales[i]);
			new = newlocale(LC_ALL_MASK, buf, NULL);
			if (new != NULL) {
				if (!strcmp(nl_langinfo_l(CODESET, new), "UTF-8")) return new;
				freelocale(new);
			}
		}
		i++;
	}
	return NULL;
}

struct language *find_or_create_language (const char *name)
{
	struct language **lp;
	struct language *l;
	lp = &first_language;
	l = first_language;
	while (l != NULL) {
		if (!strcmp(l->name, name)) return l;
		lp = &l->next;
		l = l->next;
	}
	l = malloc(sizeof(struct language));
	l->name = strdup(name);
	l->locale = NULL;
	if (name[0] == '\0') {
		name = "C";
	}
	if (!strcmp(name, "C") || !strcmp(name, "POSIX")) {
		/* cheat: call it C internally, but prefer en_US if available
		   since it has better collating rules */
		l->locale = newlocale2("en_US");
	}
	if (l->locale == NULL) {
		l->locale = newlocale2(name);
	}
	if (l->locale == NULL) {
		message(4, "locale \"%s\" not supported", name);
	}
	l->next = NULL;
	*lp = l;
	return l;
}

int section_add_value (struct section *section, const char *key,
  const char *lang, const char *value)
{
	struct key **kp = &section->keys;
	struct key *k = section->keys;
	struct lang **lp;
	struct lang *l;
	while (k != NULL) {
		if (!strcmp(k->name, key)) {
			l = k->langs;
			lp = &k->langs;
			while (l != NULL) {
				if (!strcmp(l->language->name, lang)) return -1;
				lp = &l->next;
				l = l->next;
			}
			l = malloc(sizeof(struct lang));
			l->language = find_or_create_language(lang);
			l->value = strdup(value);
			l->next = NULL;
			*lp = l;
			return 0;
		}
		kp = &k->next;
		k = k->next;
	}
	k = malloc(sizeof(struct key));
	l = malloc(sizeof(struct lang));
	l->language = find_or_create_language(lang);
	l->value = strdup(value);
	l->next = NULL;
	k->name = strdup(key);
	k->langs = l;
	k->next = NULL;
	*kp = k;
	return 0;
}

struct section *item_add_section (struct item *item, const char *section)
{
	struct section *s;
	struct section **sp;
	s = item->sections;
	sp = &item->sections;
	while (s != NULL) {
		if (!strcmp(s->name, section)) {
			return s;
		}
		sp = &s->next;
		s = s->next;
	}
	s = malloc(sizeof(struct section));
	s->name = strdup(section);
	s->keys = NULL;
	s->next = NULL;
	*sp = s;
	return s;
}

static struct lang *key_find_exact_lang (const struct key *key, const char *lang)
{
	struct lang *l;

	if (lang == NULL) lang = "";
	l = key->langs;
	while (l != NULL) {
		if (!strcmp(l->language->name, lang)) return l;
		l = l->next;
	}
	return NULL;
}

struct lang *key_find_lang (const struct key *key, const char *lang)
{
	struct lang *l;
	char *p;
	char buf[16];

	if (lang == NULL) lang = "";
	strcpy(buf, lang);
	p = strchr(buf, '_');
	if (p != NULL) {
		l = key_find_exact_lang(key, buf);
		if (l != NULL) return l;
		*p = '\0';
	}
	if (buf[0] != '\0') {
		l = key_find_exact_lang(key, buf);
		if (l != NULL) return l;
	}
	return key_find_exact_lang(key, "");
}

struct key *section_find_key (const struct section *section, const char *key)
{
	struct key *k;
	k = section->keys;
	while (k != NULL) {
		if (!strcmp(k->name, key)) return k;
		k = k->next;
	}
	return NULL;
}

struct section *item_find_section (const struct item *item, const char *section)
{
	struct section *s;
	if (section == NULL || section[0] == '\0') section = "Desktop Entry";
	s = item->sections;
	while (s != NULL) {
		if (!strcmp(s->name, section)) return s;
		s = s->next;
	}
	return NULL;
}

struct key *item_find_key (const struct item *item, const char *section,
  const char *key)
{
	struct section *s;
	s = item_find_section(item, section);
	if (s == NULL) return NULL;
	return section_find_key(s, key);
}

char *get_value (const struct item *item, const char *section,
  const char *key, const char *lang)
{
	struct section *s;
	struct key *k;
	struct lang *l;

	s = item_find_section(item, section);
	if (s == NULL) return NULL;
	k = section_find_key(s, key);
	if (k == NULL) return NULL;
	l = key_find_lang(k, lang);
	if (l == NULL) return NULL;
	return l->value;
}

int get_value_bool (const struct item *item, const char *section,
  const char *key, const char *lang)
{
	const char *s;
	s = get_value(item, section, key, lang);
	if (s == NULL || *s == '\0') return -1;
	if (!strcmp(s, "1") || !strcasecmp(s, "true") || !strcasecmp(s, "yes")) return 1;
	return 0;
}

int string_to_int (const char *s, int err)
{
	int val;
	char dummy;
	if (sscanf(s, "%d%c", &val, &dummy) != 1) return err;
	return val;
}

int is_in_list (const char *word, const char *list, const char *delim)
{
	int i, j;
	if (list == NULL) return 0;
	if (strpbrk(word, delim) != NULL) return 0;
	i = strlen(word);
	while (list[0] != '\0') {
		list += strspn(list, delim);
		j = strcspn(list, delim);
		if (i == j && !strncasecmp(list, word, i)) return 1;
		list += j;
	}
	return 0;
}

int empty_dir (const char *path)
{
	int fd;
	DIR *d;
	struct dirent *de;
	struct stat st;
	int error = 0;
	int isdir;

	fd = open(".", O_RDONLY);
	if (fd < 0) {
		return -1;
	}
	if (chdir(path) < 0) {
		close(fd);
		return -1;
	}
	if ((d = opendir(".")) == NULL) {
		fchdir(fd);
		close(fd);
		return -1;
	}
	while ((de = readdir(d)) != NULL) {
		if (!strcmp(de->d_name, "." )) continue;
		if (!strcmp(de->d_name, "..")) continue;
		isdir = 0;
		if (de->d_type == DT_DIR) {
			isdir = 1;
		} else if (de->d_type == DT_UNKNOWN) {
			if (lstat(de->d_name, &st) < 0) {
				error = 1;
				continue;
			}
			if (S_ISDIR(st.st_mode)) isdir = 1;
		}
		if (isdir) {
			if (empty_dir(de->d_name) < 0) error = 1;
			if (rmdir(de->d_name) < 0) error = 1;
		} else {
			if (unlink(de->d_name) < 0) error = 1;
		}
	}
	closedir(d);
	fchdir(fd);
	close(fd);
	return error ? -1 : 0;
}

int rm_rf (const char *path)
{
	struct stat st;
	int error = 0;

	if (lstat(path, &st) < 0) {
		return -1;
	}
	if (S_ISDIR(st.st_mode)) {
		if (empty_dir(path) < 0) error = 1;
		if (rmdir(path) < 0) error = 1;
	} else {
		if (unlink(path) < 0) error = 1;
	}
	return error ? -1 : 0;
}

static int move2 (const char *src, unsigned char srctype, const char *dst)
{
	DIR *d;
	struct dirent *de;
	char nextsrc[PATH_MAX], nextdst[PATH_MAX];
	int fd1, fd2;
	int i1, i2;
	char fd1buf[4096], fd2buf[4096];
	struct stat stsrc, stdst;
	unsigned char dsttype = DT_UNKNOWN;
	int srcstated = 0;
	int retval = 0;

	/* just one rename if "dst" doesn't exist */
	if (lstat(dst, &stdst) < 0) {
		if (errno != ENOENT) return -1;
		if (rename(src, dst) < 0) return -1;
		return 0;
	}
	if (S_ISREG(stdst.st_mode)) dsttype = DT_REG;
	if (S_ISDIR(stdst.st_mode)) dsttype = DT_DIR;
	/* see what "src" is */
	if (srctype == DT_UNKNOWN) {
		if (lstat(src, &stsrc) < 0) {
			return -1;
		}
		srcstated = 1;
		if (S_ISREG(stsrc.st_mode)) srctype = DT_REG;
		if (S_ISDIR(stsrc.st_mode)) srctype = DT_DIR;
	}

	/* if src and dst are both regular files... */
	if (srctype == DT_REG && dsttype == DT_REG) {
		if (!srcstated) {
			if (lstat(src, &stsrc) < 0) return -1;
			if (!S_ISREG(stsrc.st_mode)) return -1; /* shouldn't happen */
			srcstated = 1;
		}
		/* compare the two files and move only if they differ */
		if (stsrc.st_size != stdst.st_size) {
			return rename(src, dst);
		}
		fd1 = fd2 = -1;
		if ((fd1 = open(src, O_RDONLY)) < 0) {
			rename(src, dst);
			return -1;
		}
		if ((fd2 = open(dst, O_RDONLY)) < 0) {
			close(fd1);
			rename(src, dst);
			return -1;
		}
		while (1) {
			i1 = read(fd1, fd1buf, 4096);
			i2 = read(fd2, fd2buf, 4096);
			if (i1 < 0 || i2 < 0 || i1 != i2) {
				close(fd1);
				close(fd2);
				rename(src, dst);
				return -1;
			}
			if (i1 == 0 && i2 == 0) {
				break;
			}
			if (memcmp(fd1buf, fd2buf, i1)) {
				close(fd1);
				close(fd2);
				return rename(src, dst);
			}
		}
		close(fd1);
		close(fd2);
		return unlink(src);
	}

	/* if src and dst are both directories... */
	if (srctype == DT_DIR && dsttype == DT_DIR) {
		/* remove entries from dst that do not exist under src */
		if ((d = opendir(dst)) == NULL) {
			return -1;
		}
		while ((de = readdir(d)) != NULL) {
			if (!strcmp(de->d_name, "."))  continue;
			if (!strcmp(de->d_name, "..")) continue;
			sprintf(nextsrc, "%s/%s", src, de->d_name);
			if (lstat(nextsrc, &stsrc) < 0) {
				sprintf(nextdst, "%s/%s", dst, de->d_name);
				if (rm_rf(nextdst) < 0) retval = -1;
			}
		}
		closedir(d);
		/* handle entries of src */
		if ((d = opendir(src)) == NULL) {
			return -1;
		}
		while ((de = readdir(d)) != NULL) {
			if (!strcmp(de->d_name, "."))  continue;
			if (!strcmp(de->d_name, "..")) continue;
			sprintf(nextsrc, "%s/%s", src, de->d_name);
			sprintf(nextdst, "%s/%s", dst, de->d_name);
			if (move2(nextsrc, de->d_type, nextdst) < 0) retval = -1;
		}
		closedir(d);
		/* now src is an empty directory, we need to remove it too */
		if (rmdir(src) < 0) retval = -1;
		return retval;
	}

	/* all other cases... */
	if (rm_rf(dst) < 0) retval = -1;
	if (rename(src, dst) < 0) retval = -1;
	return retval;
}

int move (const char *src, const char *dst)
{
	return move2(src, DT_UNKNOWN, dst);
}

void keep_file (struct keep_files *keep_files, const char *file)
{
	int i;
	for (i = 0; i < keep_files->count; i++) {
		if (!strcmp(keep_files->filenames[i], file)) {
			keep_files->keep_flags[i] = 1;
			break;
		}
	}
}

void message(int level, const char *format, ...)
{
	va_list ap;

	if (level >= loglevel) return;
	fprintf(stderr, "uhu-menu");
	if (plugin_type >= 0) {
		fprintf(stderr, " [%s plugin `%s']",
		  plugin_type == 0 ? "input" : "output",
		  plugin_name);
	}
	fprintf(stderr, ": ");
	va_start(ap, format);
	vfprintf(stderr, format, ap);
	va_end(ap);
	fprintf(stderr, "\n");
}

void message_e(int level, const char *format, ...)
{
	va_list ap;
	int saved_errno = errno;

	if (level >= loglevel) return;
	fprintf(stderr, "uhu-menu");
	if (plugin_type >= 0) {
		fprintf(stderr, " [%s plugin `%s']",
		  plugin_type == 0 ? "input" : "output",
		  plugin_name);
	}
	fprintf(stderr, ": ");
	va_start(ap, format);
	vfprintf(stderr, format, ap);
	va_end(ap);
	fprintf(stderr, ": %s\n", strerror(saved_errno));
}

static int desktop_parse_line (struct item *item, struct section **sp, char *s)
{
	char *lang, *value, *v2, *v3;
	char *end = s + strlen(s);
	while (end > s && (*(end-1) == ' ' || *(end-1) == '\t')) *(--end) = '\0';
	while (*s == ' ' || *s == '\t') s++;
	if (*s == '\0' || *s == '#') return 0;
	if (*s == '[' && *(end-1) == ']') {
		*(--end) = '\0';
		*sp = item_add_section(item, s+1);
		return 0;
	}
	value = strchr(s, '=');
	if (value == NULL) return -1;
	end = value;
	*end = '\0';
	value++;
	while (*value == ' ' || *value == '\t') value++;
	while (end > s && (*(end-1) == ' ' || *(end-1) == '\t')) *(--end) = '\0';
	if (*s == '\0') return -1;
	lang = strchr(s, '[');
	if (lang != NULL && *(end-1) == ']') {
		*lang++ = '\0';
		*(--end) = '\0';
	} else {
		lang = "";
	}
	if (*sp == NULL) {
		message(1, "section == NULL !!!");
		return -1;
	}
	v2 = malloc(strlen(value) + 1);
	v3 = v2;
	while (*value != '\0') {
		if (*value == '\r') {
			value++;
			continue;
		}
		if (*value == '\\') {
			value++;
			switch (*value) {
			  case 's':
				*v3 = ' ';
				break;
			  case 'n':
				*v3 = '\n';
				break;
			  case 't':
				*v3 = '\t';
				break;
			  case 'r':
				*v3 = '\r';
				break;
			  default:
				*v3 = *value;
			}
			if (*v3 != '\0') v3++;
			value++;
		} else {
			*v3++ = *value++;
		}
	}
	*v3 = '\0';
	section_add_value(*sp, s, lang, v2);
	free(v2);
	return 0;
}

int desktop_parse (const char *file, int path_remove, struct item *item)
{
	struct stat st;
	int fd;
	char *p, *s;
	struct section *section = NULL;
	char *file_content;
	int file_size;

	fd = open(file, O_RDONLY);
	if (fd < 0) {
		message_e(3, "cannot open %s", file);
		return -1;
	}

	if (fstat(fd, &st) < 0) {
		message_e(3, "fstat");
		close(fd);
		return -1;
	}

	item->file_name = strdup(file + path_remove);
	file_size = st.st_size;
	file_content = malloc(file_size + 1);

	if (read(fd, file_content, file_size) != file_size) {
		message(3, "read failed");
		close(fd);
		free(file_content);
		return -1;
	}
	file_content[file_size] = '\0';

	close(fd);

	item->sections = NULL;
	p = file_content;
	while (p < file_content + file_size) {
		s = strchr(p, '\n');
		if (s == NULL) {
			s = p + strlen(p);
		} else {
			*s++ = '\0';
		}
		desktop_parse_line(item, &section, p);
		p = s;
	}

	free(file_content);
	return 0;
}
